"""
Ludolph: Monitoring Jabber Bot
Copyright (C) 2014-2015 Erigones, s. r. o.
This file is part of Ludolph.

See the LICENSE file for copying permission.
"""
import logging
from functools import wraps

LOG_LEVELS = frozenset(['DEBUG', 'INFO', 'WARN', 'WARNING', 'ERROR', 'FATAL', 'CRITICAL'])

logger = logging.getLogger(__name__)


def parse_loglevel(name):
    """Parse log level name and return log level integer value"""
    name = name.upper()

    if name in LOG_LEVELS:
        return getattr(logging, name, logging.INFO)

    return logging.INFO


def pluralize(count, singular, plural):
    """Return singular or plural depending on count"""
    if count == 1:
        return singular
    return plural


def catch_exception(fun):
    """
    Used as decorator to catch all exceptions and log them without breaking the inner function.
    """
    @wraps(fun)
    def wrap(*args, **kwargs):
        try:
            return fun(*args, **kwargs)
        except Exception as e:
            logger.exception(e)
            logger.error('Got exception when running %s(%s, %s): %s.', fun.__name__, args, kwargs, e)
    return wrap



"""
Ludolph: Monitoring Jabber Bot
Copyright (C) 2014-2015 Erigones, s. r. o.
This file is part of Ludolph.

See the LICENSE file for copying permission.
"""
import logging
import socket
from functools import wraps
from collections import namedtuple
# noinspection PyUnresolvedReferences
from bottle import Bottle, ServerAdapter, abort, request

__all__ = ('webhook',)

logger = logging.getLogger(__name__)


class LudolphBottle(Bottle):
    """
    Bottle web server - used for webhooks.
    """
    # noinspection PyMethodMayBeStatic
    def default_error_handler(self, res):
        return 'ERROR %s: %s\n' % (res.status_code, res.body)


WEBAPP = LudolphBottle()
WEBHOOKS = {}  # {webhook : (name, module, path)}
Webhook = namedtuple('Webhook', ('name', 'module', 'path'))


class WebServer(ServerAdapter):
    """
    Like bottle.WSGIRefServer, but with stop() method.
    """
    server = None
    quiet = True
    webhooks = WEBHOOKS

    def run(self, handler):
        logger.info('Starting web server on http://%s:%s', self.host, self.port)
        from wsgiref.simple_server import WSGIRequestHandler, WSGIServer
        from wsgiref.simple_server import make_server

        class CustomHandler(WSGIRequestHandler):
            def address_string(self):  # Prevent reverse DNS lookups
                return self.client_address[0]

            def log_error(self, *args, **kwargs):
                kwargs['level'] = logging.ERROR  # Change default log level
                self.log_message(*args, **kwargs)

            def log_message(self, fmt, *args, **kwargs):  # Log into default log file instead of stderr
                level = kwargs.get('level', logging.INFO)
                logger.log(level, '%s - - %s', self.client_address[0], str(fmt % args).rstrip('\n'))

        handler_cls = self.options.get('handler_class', CustomHandler)
        server_cls = self.options.get('server_class', WSGIServer)

        if ':' in self.host:  # Fix wsgiref for IPv6 addresses
            if getattr(server_cls, 'address_family') == socket.AF_INET:
                # noinspection PyPep8Naming
                class server_cls(server_cls):
                    address_family = socket.AF_INET6

        self.server = make_server(self.host, self.port, handler, server_cls, handler_cls)
        self.server.serve_forever()

    def stop(self):
        assert self.server, 'Web server was not started?'
        logger.info('Stopping web server')

        if self.server:
            self.server.shutdown()

    def start(self):
        assert self.server is None, 'Web server is already running?'
        global WEBAPP
        WEBAPP.run(server=self)

    def reset_webhooks(self, module=None):
        if module:
            logger.info('Deregistering webhooks from plugin: %s', module)
            global WEBAPP

            for name, hook in tuple(self.webhooks.items()):  # Copy for python 3
                if hook.module == module:
                    logger.debug('Deregistering webhook "%s" from plugin "%s" mapped to URL "%s"',
                                 name, hook.module, hook.path)
                    del self.webhooks[name]
        else:
            logger.info('Reinitializing webhooks')
            self.webhooks.clear()

    def reset_webapp(self):
        if self.server:
            logger.info('Reinitializing web server')
            global WEBAPP
            del WEBAPP
            WEBAPP = LudolphBottle()
            self.server.set_app(WEBAPP)

    def display_webhooks(self):
        """Return list of available webhooks suitable for logging"""
        return ['%s [%s]: %s' % (name, hook.module, hook.path) for name, hook in self.webhooks.items()]


def _webview(fun):
    """
    Wrapper for bottle callbacks responsible for finding back the bound method. Inspired by err bot.
    """
    @wraps(fun)
    def wrap(*args, **kwargs):
        from ludolph.bot import PLUGINS

        try:
            obj = PLUGINS[WEBHOOKS[fun.__name__].module]
            obj_fun = getattr(obj, fun.__name__)
        except (KeyError, AttributeError) as e:
            logger.error('Requested webhook "%s" is not registered (%s)', fun.__name__, e)
            abort(404, 'Webhook vanished')
        else:
            return obj_fun(*args, **kwargs)

    return wrap


def webhook(path, methods=('GET',)):
    """
    Decorator for registering HTTP request handlers. Inspired by err bot.
    """
    def webhook_decorator(fun):
        if fun.__name__ in WEBHOOKS:
            logger.critical('Webhook "%s" from plugin "%s" overlaps with existing webhook from module "%s"',
                            fun.__name__, fun.__module__, WEBHOOKS[fun.__name__].module)
            return None

        logger.debug('Registering webhook "%s" from plugin "%s" to URL "%s"', fun.__name__, fun.__module__, path)
        WEBAPP.route(path, methods, _webview(fun), name=fun.__name__)
        WEBHOOKS[fun.__name__] = Webhook(fun.__name__, fun.__module__, path)

        return fun

    return webhook_decorator



"""
Ludolph: Monitoring Jabber bot
Copyright (C) 2012-2015 Erigones, s. r. o.
This file is part of Ludolph.

See the file LICENSE for copying permission.
"""
import time
import logging
import os
import imghdr
from datetime import datetime, timedelta
from sleekxmpp.exceptions import XMPPError, IqError
from glob import iglob

# noinspection PyPep8Naming
from ludolph import __doc__ as ABOUT
from ludolph import __version__
from ludolph.command import CommandError, MissingParameter, command
from ludolph.web import webhook, request, abort
from ludolph.utils import pluralize
from ludolph.plugins.plugin import LudolphPlugin

logger = logging.getLogger(__name__)


class Base(LudolphPlugin):
    """
    Ludolph jabber bot base commands.
    """
    __version__ = __version__
    _avatar_allowed_extensions = frozenset(['.png', '.jpg', '.jpeg', '.gif'])
    _status_show_types = frozenset(['online', 'away', 'chat', 'dnd', 'xa'])  # online is a fake type translated to None
    _help_cache = None
    _cron_required = ('at', 'remind')
    _reminder = '__You have asked me to remind you:__ '

    def __post_init__(self):
        # Disable at command if cron is disabled
        if not self.xmpp.cron:
            for i in self._cron_required:
                self.xmpp.commands.pop(i)

        # Reset help command cache
        self._help_cache = None
        # Override fallback message handler
        self.xmpp.fallback_message = self._fallback_message

    def __destroy__(self):
        # Recover original fallback message handler
        self.xmpp.fallback_message = self.xmpp.original_fallback_message

    def _fallback_message(self, msg, cmd_name):
        """Fallback message handler called in case the command does not exist"""
        self.xmpp.msg_reply(msg, 'Sorry, I don\'t understand __"%s"__\n'
                                 'Please type **help** for more info' % cmd_name)

    def _help_all(self):
        """Return list of all commands organized by plugins"""
        if self._help_cache is None:
            # Create dict with module name as key and list of commands as value
            xmpp = self.xmpp
            cmd_map = {}

            for cmd_name in xmpp.commands.all():
                cmd = xmpp.commands[cmd_name]
                mod_name = cmd.module

                if mod_name not in cmd_map:
                    cmd_map[mod_name] = []

                cmd_map[mod_name].append(cmd)

            out = ['List of available **%s** commands:' % xmpp.nick]

            for mod_name, plugin in xmpp.plugins.items():  # The plugins dict knows the plugin order
                try:
                    commands = cmd_map[mod_name]
                except KeyError:
                    continue

                # Item: module name
                if plugin.__version__:
                    version = '^^%s^^' % plugin.__version__
                else:
                    version = ''

                out.append('\n* %s %s\n' % (mod_name, version))

                for cmd in commands:
                    try:
                        # First line of __doc__
                        desc = cmd.doc.split('\n')[0].replace(xmpp.__class__.nick, xmpp.nick)
                        # Lowercase first char and remove trailing dot
                        desc = ' - ' + desc[0].lower() + desc[1:].rstrip('.')
                    except IndexError:
                        desc = ''

                    # SubItem: line of command + description
                    out.append('  * **%s**%s' % (cmd.name, desc))

            out.append('\nUse "help <command>" for more information about the command usage')
            self._help_cache = '\n'.join(out)

        return self._help_cache

    # noinspection PyUnusedLocal
    @command
    def help(self, msg, cmdstr=None):
        """
        Show this help.

        Usage: help [command]
        """
        # Global help or command help?
        if cmdstr:
            xmpp = self.xmpp
            cmd = xmpp.commands.get_command(cmdstr)

            if cmd:
                # Remove whitespaces from __doc__ lines
                desc = '\n'.join(map(str.strip, cmd.doc.split('\n')))
                # **Command name** (module) + desc
                return '**%s** (%s)\n\n%s' % (cmd.name, cmd.module, desc.replace(xmpp.__class__.nick, xmpp.nick))

        return self._help_all()

    # noinspection PyUnusedLocal
    @command
    def version(self, msg, plugin=None):
        """
        Display version of Ludolph or registered plugin.

        Usage: version [plugin]
        """
        if plugin:
            mod, obj = self.xmpp.plugins.get_plugin(plugin)

            if mod:
                return '**%s** version: %s' % (mod, obj.get_version())
            else:
                raise CommandError('**%s** isn\'t a Ludolph plugin. Check help for available plugins.' % plugin)

        return '**Ludolph** version: %s' % self.get_version()

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    @command
    def about(self, msg):
        """
        Details about this project.

        Usage: about
        """
        return ABOUT.strip()

    # noinspection PyUnusedLocal
    @command
    def uptime(self, msg):
        """
        Show Ludolph uptime.

        Usage: uptime
        """
        # noinspection PyProtectedMember
        u = time.time() - self.xmpp._start_time
        m, s = divmod(u, 60)
        h, m = divmod(m, 60)
        d, h = divmod(h, 24)

        return 'up %d days, %d hours, %d minutes, %d seconds' % (d, h, m, s)

    def _message_send(self, jid, msg):
        """Send new xmpp message. Used by message command and /message webhook"""
        if jid == self.xmpp.room:
            mtype = 'groupchat'
        elif jid in self.xmpp.client_roster:
            mtype = 'normal'
        else:
            raise CommandError('User "%s" not in roster' % jid)

        logger.info('Sending message to "%s"', jid)
        logger.debug('\twith body: "%s"', msg)
        self.xmpp.msg_send(jid, msg, mtype=mtype)

        return 'Message sent to **%s**' % jid

    # noinspection PyUnusedLocal
    @command
    def message(self, msg, jid, text):
        """
        Send new XMPP message to user/room.

        Usage: message <JID> <text>
        """
        return self._message_send(jid, text)

    # noinspection PyUnusedLocal
    @command(admin_required=True)
    def broadcast(self, msg, text):
        """
        Send private message to every user in roster (admin only).

        Usage: broadcast <message>
        """
        return 'Message broadcasted to %dx users.' % self.xmpp.msg_broadcast(text)

    def _set_status(self, show, status=None):
        """Send presence status"""
        if show not in self._status_show_types:
            raise CommandError('Invalid status type')

        if show == 'online':
            show = None

        try:
            self.xmpp.send_presence(pstatus=status, pshow=show)
        except IqError as e:
            raise CommandError('Status update failed: __%s__' % getattr(e, 'condition', str(e)))

    # noinspection PyUnusedLocal
    @command(admin_required=True)
    def status(self, msg, show, status=None):
        """
        Set Ludolph's status (admin only).

        Usage: status {online|away|chat|dnd|xa} [status message]
        """
        self._set_status(show, status=status)

        return 'Status updated'

    def _roster_list(self):
        """List users on Ludolph's roster (admin only)"""
        roster = self.xmpp.client_roster

        return '\n'.join(['%s\t%s' % (i, roster[i]['subscription']) for i in roster])

    def _roster_del(self, user):
        """Remove user from Ludolph's roster (admin only)"""
        if user in self.xmpp.client_roster:
            self.xmpp.send_presence(pto=user, ptype='unsubscribe')
            self.xmpp.del_roster_item(user)

            return 'User **%s** removed from roster' % user
        else:
            return 'User **%s** cannot be removed from roster' % user

    # noinspection PyUnusedLocal
    @command(admin_required=True)
    def roster(self, msg, action=None, user=None):
        """
        List and manage users on Ludolph's roster (admin only).

        List users on Ludolph's roster.
        Usage: roster

        Remove user from Ludolph's roster
        Usage: roster del <JID>
        """
        if action == 'del':
            if user:
                return self._roster_del(user)
            else:
                raise MissingParameter

        return self._roster_list()

    def _get_avatar_dirs(self):
        """Get list of directories where avatars are stored."""
        avatar_dir = self.config.get('avatar_dir', None)
        default_avatar_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'avatars')

        if avatar_dir:
            return avatar_dir, default_avatar_dir
        else:
            return default_avatar_dir,

    def _avatar_list(self):
        """List available avatars for Ludolph (admin only)"""
        files = []

        for avatar_dir in self._get_avatar_dirs():
            if os.path.isdir(avatar_dir):
                for file_type in self._avatar_allowed_extensions:
                    for avatar in iglob(os.path.join(avatar_dir, '*' + file_type)):
                        files.append(os.path.basename(avatar))
            else:
                logger.warning('Avatars directory: %s does not exists.' % avatar_dir)
                continue

        if files:
            return 'List of available avatars: %s' % ', '.join(files)
        else:
            return 'No avatars were found... :('

    def _avatar_set(self, msg, avatar_name):
        """Set avatar for Ludolph (admin only)"""
        if os.path.splitext(avatar_name)[-1] not in self._avatar_allowed_extensions:
            raise CommandError('You have requested a file that is not supported')

        avatar = None
        available_avatar_directories = self._get_avatar_dirs()

        for avatar_dir in available_avatar_directories:
            # Create full path to file requested by user
            avatar_file = os.path.join(avatar_dir, avatar_name)
            # Split absolute path for check if user is not trying to jump outside allowed dirs
            path, name = os.path.split(os.path.abspath(avatar_file))

            if path not in available_avatar_directories:
                raise CommandError('You are not allowed to set avatar outside defined directories')

            try:
                with open(avatar_file, 'rb') as f:
                    avatar = f.read()
            except (OSError, IOError):
                avatar = None
            else:
                break

        if not avatar:
            raise CommandError('Avatar "%s" has not been found.\n'
                               'You can list available avatars with the command: **avatar-list**' % avatar_name)

        self.xmpp.msg_reply(msg, 'I have found the selected avatar, changing it might take few seconds...',
                            preserve_msg=True)
        avatar_type = 'image/%s' % imghdr.what('', avatar)
        avatar_id = self.xmpp.plugin['xep_0084'].generate_id(avatar)
        avatar_bytes = len(avatar)

        try:
            logger.debug('Publishing XEP-0084 avatar data')
            self.xmpp.plugin['xep_0084'].publish_avatar(avatar)
        except XMPPError as e:
            logger.error('Could not publish XEP-0084 avatar: %s' % e.text)
            raise CommandError('Could not publish selected avatar')

        try:
            logger.debug('Publishing XEP-0153 avatar vCard data')
            self.xmpp.plugin['xep_0153'].set_avatar(avatar=avatar, mtype=avatar_type)
        except XMPPError as e:
            logger.error('Could not publish XEP-0153 vCard avatar: %s' % e.text)
            raise CommandError('Could not set vCard avatar')

        self.xmpp.msg_reply(msg, 'Almost done, please be patient', preserve_msg=True)

        try:
            logger.debug('Advertise XEP-0084 avatar metadata')
            self.xmpp['xep_0084'].publish_avatar_metadata([{
                'id': avatar_id,
                'type': avatar_type,
                'bytes': avatar_bytes
            }])
        except XMPPError as e:
            logger.error('Could not publish XEP-0084 metadata: %s' % e.text)
            raise CommandError('Could not publish avatar metadata')

        return 'Avatar has been changed :)'

    @command(admin_required=True)
    def avatar(self, msg, action=None, avatar_name=None):
        """
        List available avatars or set an avatar for Ludolph (admin only).

        List available avatars for Ludolph.
        Usage: avatar

        Set avatar for Ludolph.
        Usage: avatar set <avatar>
        """
        if action == 'set':
            if avatar_name:
                return self._avatar_set(msg, avatar_name)
            else:
                raise MissingParameter

        return self._avatar_list()

    def _at_list(self, msg, reminder=False):
        """List all scheduled jobs"""
        crontab = self.xmpp.cron.crontab
        user = self.xmpp.get_jid(msg)

        if reminder:
            display_job = lambda cronjob: cronjob.onetime and user == cronjob.owner \
                and cronjob.command.split(' ')[:2] == ['message', user]

            out = ['**%s** [%s] __%s__' % (name, job.schedule,
                                           ' '.join(job.command.split(' ')[2:]).replace(self._reminder + ' ', ''))
                   for name, job in crontab.items() if display_job(job)]
        else:

            if self.xmpp.is_jid_admin(user):
                display_job = lambda cronjob: cronjob.onetime
            else:
                display_job = lambda cronjob: cronjob.onetime and user == cronjob.owner

            out = ['**%s** [%s] (%s) __%s__' % (name, job.schedule, job.owner, job.command)
                   for name, job in crontab.items() if display_job(job)]

        count = len(out)
        out.append('\n**%d** %s scheduled' % (count, pluralize(count, 'job is', 'jobs are')))

        return '\n'.join(out)

    def _at_del(self, msg, name):
        """Remove scheduled job"""
        try:
            job_id = int(name)
        except (ValueError, TypeError):
            raise CommandError('Invalid job ID')

        crontab = self.xmpp.cron.crontab
        job = crontab.get(job_id, None)

        if job and job.onetime:
            user = self.xmpp.get_jid(msg)

            if job.owner == user or self.xmpp.is_jid_admin(user):
                crontab.delete(job_id)
                logger.info('Deleted one-time cron jobs: %s', job.display())

                return 'Scheduled job ID **%s** deleted' % job_id
            else:
                raise CommandError('Permission denied')

        raise CommandError('Non-existent job ID')

    def _at_add(self, msg, schedule, cmd_name, *cmd_args, **job_kwargs):
        """Schedule command execution at specific time and date"""
        # Validate schedule
        schedule = str(schedule)

        if schedule.startswith('+'):
            try:
                dt = datetime.now() + timedelta(minutes=int(schedule))
            except ValueError:
                raise CommandError('Invalid date-time (required format: +<integer>)')
        else:
            try:
                dt = datetime.strptime(schedule, '%Y-%m-%d-%H-%M')
            except ValueError:
                raise CommandError('Invalid date-time (required format: YYYY-mm-dd-HH-MM)')

        # Validate command
        cmd = self.xmpp.commands.get_command(cmd_name)

        if not cmd:
            raise CommandError('Invalid command')

        # Check user permission
        user = self.xmpp.get_jid(msg)

        if not cmd.is_jid_permitted_to_run(self.xmpp, user):
            raise CommandError('Permission denied')

        # Create message (the only argument needed for command) with body representing the whole command
        body = ' '.join([cmd.name] + ["%s" % i for i in cmd_args])
        msg = self.xmpp.msg_copy(msg, body=body)
        job = self.xmpp.cron.crontab.add_at(cmd.get_fun(self.xmpp), dt, msg, user, **job_kwargs)
        logger.info('Registered one-time cron job: %s', job.display())

        return 'Scheduled job ID **%s** scheduled at %s' % (job.name, job.schedule)

    @command
    def at(self, msg, *args):
        """
        List, add, or delete jobs for later execution.

        List all scheduled jobs.
        Usage: at

        Schedule command execution at specific time and date.
        Usage: at add +minutes <command> [command parameters...]
        Usage: at add Y-m-d-H-M <command> [command parameters...]

        Remove command from queue of scheduled jobs.
        Usage: at del <job ID>
        """
        if not self.xmpp.cron:
            raise CommandError('Cron support is disabled in Ludolph configuration file')

        args_count = len(args)

        if args_count > 0:
            action = args[0]

            if action == 'add':
                if args_count < 3:
                    raise MissingParameter
                else:
                    return self._at_add(msg, *args[1:])
            elif action == 'del':
                if args_count < 2:
                    raise MissingParameter
                else:
                    return self._at_del(msg, args[1])
            else:
                raise CommandError('Invalid action')

        return self._at_list(msg)

    @command
    def remind(self, msg, *args):
        """
        List, add, or delete reminders.

        List all scheduled reminders.
        Usage: remind

        Schedule reminder at specific time and date.
        Usage: remind add +minutes <message>
        Usage: remind add Y-m-d-H-M <message>

        Remove reminder from queue of scheduled reminders.
        Usage: remind del <reminder ID>
        """
        args_count = len(args)

        if args_count > 0:
            action = args[0]

            if action == 'add':
                if args_count < 2:
                    raise MissingParameter
                else:
                    return self._at_add(msg, *(args[1], 'message', self.xmpp.get_jid(msg), self._reminder) + args[2:],
                                        at_reply_output=False)
            elif action == 'del':
                if args_count < 2:
                    raise MissingParameter
                else:
                    return self._at_del(msg, args[1])
            else:
                raise CommandError('Invalid action')

        return self._at_list(msg, reminder=True)

    @webhook('/')
    def index(self):
        """
        Default web page.
        """
        return ABOUT

    @webhook('/ping')
    def ping(self):
        """
        Ping-pong.
        """
        return 'pong'

    @webhook('/message', methods=('POST',))
    def send_msg(self):
        """
        Send xmpp message to user/room.
        """
        jid = request.forms.get('jid', None)

        if not jid:
            abort(400, 'Missing JID in message request')

        msg = request.forms.get('msg', '')

        try:
            return self._message_send(jid, msg)
        except CommandError as e:
            abort(400, str(e))

    @webhook('/broadcast', methods=('POST',))
    def broadcast_msg(self):
        """
        Send private message to every user in roster.
        """
        msg = request.forms.get('msg', None)

        if not msg:
            logger.warning('Missing msg parameter in broadcast request')
            abort(400, 'Missing msg parameter')

        return 'Message sent (%dx)' % self.xmpp.msg_broadcast(msg)



"""
Ludolph: Monitoring Jabber bot
Copyright (C) 2015 Erigones, s. r. o.
This file is part of Ludolph.

See the file LICENSE for copying permission.
"""
import logging
from shelve import Shelf

try:
    import anydbm as dbm
except ImportError:
    import dbm

logger = logging.getLogger(__name__)

__all__ = ('LudolphDB', 'LudolphDBMixin')


class LudolphDB(Shelf):
    """
    Dictionary-like object used for saving/loading persistent data.
    """
    def __init__(self, filename, flag='c', protocol=None, writeback=False):
        self.filename = filename
        logger.info('Opening persistent DB file %s', filename)
        Shelf.__init__(self, dbm.open(filename, flag, mode=0o600), protocol, writeback)
        # logger.debug('Persistent DB file %s loaded following items: %s', filename, self)

    def __setitem__(self, key, value):
        logger.debug('Assigning item %r to persistent DB key "%s"', value, key)
        Shelf.__setitem__(self, key, value)

    def __delitem__(self, key):
        logger.debug('Removing key "%s" from persistent DB', key)
        Shelf.__delitem__(self, key)

    def sync(self):
        logger.info('Syncing persistent DB file %s', self.filename)
        Shelf.sync(self)
        # logger.debug('Persistent DB file %s synced with following items: %s', self.filename, self)

    def close(self):
        logger.info('Closing persistent DB file %s', self.filename)
        Shelf.close(self)
        # logger.debug('Persistent DB file %s closed with following items: %s', self.filename, self)


class LudolphDBMixin(object):
    """
    Interface for classes that want to use the LudolphDB object.
    """
    db = None

    def __init__(self, db=None):
        """Enable DB support if available"""
        if db is not None:
            self.db_enable(db, init=True)

    def _db_set_items(self):
        """Set/associate some object(s) with some DB key"""
        raise NotImplementedError

    def _db_load_items(self):
        """Load data from DB and update your object(s)"""
        raise NotImplementedError

    def db_enable(self, db, init=False):
        """Enable DB support in your object"""
        self.db = db

        if db is not None:
            if init:
                self._db_load_items()

            self._db_set_items()

    def db_disable(self):
        """Disable DB support in your object"""
        self.db = None



"""
Ludolph: Monitoring Jabber bot
Copyright (C) 2012-2015 Erigones, s. r. o.
This file is part of Ludolph.

See the file LICENSE for copying permission.
"""


class LudolphPlugin(object):
    """
    Ludolph plugin base class.
    """
    __version__ = None
    persistent_attrs = ()  # Set of object's attributes that will be saved/loaded during bot's shutdown/start events.

    # noinspection PyUnusedLocal
    def __init__(self, xmpp, config, reinit=False, **kwargs):
        self.xmpp = xmpp  # Reference to LudolphBot object
        self.config = dict(config)  # Plugin configuration as list of (name, value) tuples
        self._reloaded = reinit

    def __repr__(self):
        return '<LudolphPlugin: %s.%s>' % (self.__class__.__module__, self.__class__.__name__)

    # noinspection PyMethodMayBeStatic
    def __post_init__(self):
        """Run after ludolph bot instance is up and running"""
        pass

    # noinspection PyMethodMayBeStatic
    def __destroy__(self):
        """Run before ludolph bot reload or shutdown"""
        pass

    def __getstate__(self):
        # FIXME: Switch to dict comprehension after dropping support for Python 2.6
        return dict((i, self.__dict__[i]) for i in self.persistent_attrs if i in self.__dict__)

    def __setstate__(self, state):
        for i in state:
            if i in self.persistent_attrs:
                self.__dict__[i] = state[i]

    def _db_save(self):
        """Save persistent attributes now"""
        if self.xmpp.db is not None:
            # noinspection PyProtectedMember
            self.xmpp._db_set_item(self.__class__.__module__, self)

    def _db_load(self):
        """Load persistent attributes from DB"""
        if self.xmpp.db is not None:
            # noinspection PyProtectedMember
            self.xmpp._db_load_item(self.__class__.__module__, self)

    @classmethod
    def get_version(cls):
        """Used by the version command"""
        if cls.__version__ is None:
            return '**Not implemented** by plugin author...'
        else:
            return '**%s**' % cls.__version__



"""
Ludolph: Monitoring Jabber Bot
Version: 0.7.0
Homepage: https://github.com/erigones/Ludolph
Copyright (C) 2012-2015 Erigones, s. r. o.

See the LICENSE file for copying permission.
"""

__version__ = '0.7.0'



"""
Ludolph: Monitoring Jabber Bot
Original Library: Copyright (C) 2011 JS Lee (https://github.com/jsleetw/crontab.py)
Ludolph Modification: Copyright (C) 2014-2015 Erigones, s. r. o.
This file is part of Ludolph.

See the LICENSE file for copying permission.
"""
import logging
import time
from datetime import datetime, timedelta
from functools import wraps
from collections import namedtuple

try:
    from collections import OrderedDict
except ImportError:
    # noinspection PyUnresolvedReferences,PyPackageRequirements
    from ordereddict import OrderedDict

from ludolph.message import IncomingLudolphMessage
from ludolph.db import LudolphDBMixin

__all__ = ('cronjob',)

logger = logging.getLogger(__name__)

CronJobFun = namedtuple('CronJobFun', ('name', 'module'))


def at_fun(job, fun):
    """
    Decorator for "at" job command functions.
    """
    @wraps(fun)
    def wrap(msg, *args, **kwargs):
        msg.reply_output = False  # Do not send command output to job owner
        ret = fun(msg, *args, **kwargs)

        out = 'Scheduled job **%s** run at %s finished with output:\n%s' % (job.name, datetime.now().isoformat(), ret)

        if job.at_reply_output:
            fun.__self__.xmpp.msg_reply(msg, out)  # We will inform the owner here

        return out
    return wrap


class CronStar(set):
    """Universal set - match everything"""
    def __contains__(self, item):
        return True
star = CronStar()


class CronJobError(Exception):
    pass


class CronJob(object):
    """
    Crontab entry.
    """
    def __init__(self, name, fun, args=(), kwargs=(), minute=None, hour=None, day=None, month=None, dow=None,
                 onetime=False, owner=None, at=False, at_reply_output=False):
        if not isinstance(fun, CronJobFun):
            raise TypeError('fun must be a instance of CronJobFun')

        self.name = name
        self._fun = fun
        self.args = args
        self.kwargs = dict(kwargs)
        self.minutes = self.validate_field(minute, 0, 59)
        self.hours = self.validate_field(hour, 0, 23)
        self.days = self.validate_field(day, 1, 31)
        self.months = self.validate_field(month, 1, 12)
        self.dow = self.validate_field(dow, 0, 6)
        self.onetime = self.clean_datetime(onetime)
        self.owner = owner
        self.at = at
        self.at_reply_output = at_reply_output

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.name)

    @property
    def schedule(self):
        """String representation of cron/onetime schedule"""
        if self.onetime:
            return self.onetime.isoformat()

        j = lambda x: ','.join(map(str, x)) or '*'
        return '%s %s %s %s %s' % (j(self.minutes), j(self.hours), j(self.days), j(self.months), j(self.dow))

    @property
    def module(self):
        """Return plugin name"""
        return self._fun.module

    @property
    def fqfn(self):
        """Fully qualified function name"""
        return '%s.%s' % (self._fun.module, self._fun.name)

    @property
    def command(self):
        """String representation of job command with all arguments"""
        if self.at and self.args:
            return self.args[0].get('body')

        cmd = [self._fun.name.replace('_', '-')]

        if self.args:
            cmd.extend(map(str, self.args))

        if self.kwargs:
            cmd.extend(['%s=%s' % kv for kv in self.kwargs.items()])

        return ' '.join(cmd)

    @property
    def fun(self):
        """Get the real fun - the plugin object's bound method"""
        from ludolph.bot import PLUGINS

        try:
            obj = PLUGINS[self._fun.module]
            obj_fun = getattr(obj, self._fun.name)
        except (KeyError, AttributeError):
            raise CronJobError('%r lost its fun' % self)

        if self.at:
            obj_fun = at_fun(self, obj_fun)

        return obj_fun

    @staticmethod
    def clean_datetime(dt):
        """Return datetime object without seconds"""
        if dt:
            return datetime(*dt.timetuple()[:5])
        return None

    def display(self):
        """Return string representation of this cron job suitable for logging"""
        return '%s: %s [%s]: %s' % (self.name, self.command, self.module, self.schedule)

    # noinspection PyMethodMayBeStatic
    def validate_value(self, value, min_value, max_value):
        """Each date/time field must be a number in specified range"""
        num = int(value)

        if num > max_value or num < min_value:
            raise ValueError('Value not in range')

        return num

    def validate_field(self, value, min_value, max_value):
        """Simple validation of crontab fields"""
        if value is None or value == '*':
            return star

        if isinstance(value, int):
            value = (value,)
        elif not (value and getattr(value, '__iter__', False)):
            raise ValueError('Invalid date/time field')

        validate = self.validate_value

        return set([validate(i, min_value, max_value) for i in value])

    def match_time(self, dt):
        """Return True if this event should trigger at the specified datetime"""
        if self.onetime:
            return self.onetime <= dt
        return ((dt.minute in self.minutes) and (dt.hour in self.hours) and (dt.day in self.days) and
                (dt.month in self.months) and (dt.weekday() in self.dow))

    def run(self):
        """Go!"""
        fun = self.fun

        try:
            if self.at:
                return fun(IncomingLudolphMessage.load(self.args[0]))
            else:
                return fun(*self.args, **self.kwargs)
        except CronJobError as e:
            logger.error('Error while running cron job "%s" (%s): %s', self.name, self.fqfn, e)
            return None


class CronTab(OrderedDict):
    """
    "List" of crontab entries. Each entry is identified by a unique name.
    """
    db = None

    # noinspection PyMethodOverriding
    def __repr__(self):
        return '%s(jobs=%s)' % (self.__class__.__name__, len(self))

    def __setitem__(self, key, value, **kwargs):
        if not isinstance(value, CronJob):
            raise TypeError('value must be a instance of CronJob')

        return super(CronTab, self).__setitem__(key, value, **kwargs)

    def sync(self):
        """Store only onetime cron jobs into persistent DB"""
        try:
            if self.db is not None:
                self.db['crontab'] = self.__class__((name, job) for name, job in self.items() if job.onetime)
        except Exception as ex:
            logger.exception(ex)
            logger.critical('Could not sync crontab with persistent DB file')

    def load(self):
        """Load cronjobs from external source"""
        try:
            if self.db is not None:
                cronjobs = self.db.get('crontab', None)

                if cronjobs:
                    logger.info('Loading %d cron job(s) from persistent DB file', len(cronjobs))
                    self.update(cronjobs)
        except Exception as ex:
            logger.exception(ex)
            logger.critical('Could not load crontab from persistent DB file')

    def add(self, name, fun, **kwargs):
        """Add named crontab entry, which will run fun at specified date/time"""
        if name in self:
            raise NameError('Cron job with name "%s" is already defined' % name)

        if not hasattr(fun, '__call__'):
            raise ValueError('fun must be a callable function')

        job = CronJob(name, CronJobFun(fun.__name__, fun.__module__), **kwargs)
        self[name] = job

        if job.onetime:
            self.sync()

        return job

    def delete(self, name):
        """Delete named crontab entry"""
        job = self.pop(name)

        if job.onetime:
            self.sync()

        return job

    def generate_id(self):
        """Generate new job ID for a new onetime ("at") job"""
        try:
            last = int(list(self.keys())[-1])
        except (IndexError, ValueError):
            return 1
        else:
            return last + 1

    def add_onetime(self, fun, onetime, **kwargs):
        """Add onetime job into crontab"""
        kwargs['onetime'] = onetime

        return self.add(self.generate_id(), fun, **kwargs)

    def add_at(self, fun, onetime, msg, owner, at_reply_output=True):
        """Add "at" onetime job into crontab"""
        return self.add_onetime(fun, onetime, args=(msg.dump(),), owner=owner, at=True, at_reply_output=at_reply_output)

    def clear_cron_jobs(self):
        """Remove all cron jobs, but keep onetime jobs"""
        for name, job in tuple(self.items()):  # Copy for python 3
            if not job.onetime:
                del self[name]

    def display_cron_jobs(self):
        """Return list of available non-onetime cron jobs suitable for logging"""
        return (job.display() for job in self.values() if not job.onetime)


CRONJOBS = CronTab()


class Cron(LudolphDBMixin):
    """
    Cron thread (the scheduler).
    """
    _running = False
    running = False
    crontab = CRONJOBS

    def _db_set_items(self):
        self.crontab.sync()

    def _db_load_items(self):
        self.crontab.load()

    def db_enable(self, db, init=False):
        self.crontab.db = db
        super(Cron, self).db_enable(db, init=init)

    def db_disable(self):
        self.crontab.db = None
        super(Cron, self).db_disable()

    def run(self):
        assert not self.running, 'Cron is already running?'
        logger.info('Starting cron')
        dt = CronJob.clean_datetime(datetime.now())
        self.running = self._running = True

        try:
            while self._running:
                for name, job in tuple(self.crontab.items()):  # Copy for python 3
                    if not self._running:
                        break

                    if job.match_time(dt):
                        logger.info('Running cron job "%s" (%s) with schedule "%s" as user "%s"',
                                    name, job.fqfn, job.schedule, job.owner)

                        try:
                            res = job.run()
                        except Exception as ex:
                            logger.exception(ex)
                            logger.critical('Error while running cron job "%s" (%s)', name, job.fqfn)
                            continue
                        finally:
                            if job.onetime:
                                self.crontab.delete(name)

                        logger.info('Cron job "%s" (%s) output: "%s"', name, job.fqfn, res)

                dt += timedelta(minutes=1)

                while self._running and datetime.now() < dt:
                    time.sleep(1)

        finally:
            self.running = False

    def stop(self):
        assert self.running, 'Cron was not started?'
        logger.info('Stopping cron')
        self._running = False

        while self.running:  # Wait for running job to finish
            time.sleep(0.5)

        self.db_disable()
        logger.debug('Cron stopped')

    def reset(self, module=None):
        if module:
            logger.info('Deregistering cron jobs from plugin: %s', module)

            for name, job in tuple(self.crontab.items()):  # Copy for python 3
                if job.module == module:
                    logger.debug('Deregistering cron job "%s" from plugin "%s"', name, job.module)
                    self.crontab.delete(name)
        else:
            logger.info('Reinitializing crontab')
            self.crontab.clear_cron_jobs()

    def display_cronjobs(self):
        return self.crontab.display_cron_jobs()


def cronjob(minute='*', hour='*', day='*', month='*', dow='*'):
    """
    Decorator for creating crontab entries.
    """
    def cronjob_decorator(fun):
        if fun.__name__ in CRONJOBS:
            logger.critical('Cron job "%s" from plugin "%s" overlaps with existing cron job from module "%s"',
                            fun.__name__, fun.__module__, CRONJOBS[fun.__name__].fun.__module__)
            return None

        logger.debug('Registering cron job "%s" from plugin "%s" to run at "%s %s %s %s %s"',
                     fun.__name__, fun.__module__, minute, hour, day, month, dow)
        CRONJOBS.add(fun.__name__, fun, minute=minute, hour=hour, day=day, month=month, dow=dow)

        return fun

    return cronjob_decorator



"""
Ludolph: Monitoring Jabber Bot
Copyright (C) 2012-2015 Erigones, s. r. o.
This file is part of Ludolph.

See the LICENSE file for copying permission.
"""
from logging import getLogger
from functools import wraps
from collections import namedtuple
import shlex
import inspect

__all__ = ('CommandError', 'PermissionDenied', 'MissingParameter', 'command')

logger = getLogger(__name__)


class CommandError(Exception):
    """
    Send a standard error message to user.
    """
    error_message = ''

    def __init__(self, msg=None):
        self.error_message = msg or self.error_message

    def __str__(self):
        return 'ERROR: %s' % self.error_message


class PermissionDenied(CommandError):
    error_message = 'Permission denied'


class MissingParameter(CommandError):
    error_message = 'Missing parameter'


CommandPermissions = namedtuple('CommandPermissions', ('user_required', 'admin_required', 'room_user_required',
                                                       'room_admin_required'))

CommandParameters = namedtuple('CommandParameters', ('args_count', 'kwargs_count', 'star_args'))


class Command(namedtuple('Command', ('name', 'fun_name', 'module', 'doc', 'perms', 'fun_spec'))):
    """
    Ludolph command wrapper.
    """
    def __str__(self):
        return '%s.%s' % (self.module, self.fun_name)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self)

    def get_fun(self, bot):
        """Get command bound method from plugin"""
        return getattr(bot.plugins[self.module], self.fun_name)

    def is_jid_permitted_to_run(self, xmpp, jid):
        """Return True if user is allowed to run the command"""
        perms = self.perms

        for perm, check in ((perms.user_required, xmpp.is_jid_user),
                            (perms.admin_required, xmpp.is_jid_admin),
                            (perms.room_user_required, xmpp.is_jid_room_user),
                            (perms.room_admin_required, xmpp.is_jid_room_admin)):
            if perm and not check(jid):
                return False

        return True

    def get_args_from_msg_body(self, body):
        """Parse message body and return a list which can be used as *args parameter for this command"""
        fun_spec = self.fun_spec
        last_pos = fun_spec.args_count + fun_spec.kwargs_count - 1

        if last_pos < 0 and not fun_spec.star_args:  # Function has no custom arguments
            return []

        # Try to get command parameters (with command name removed)
        try:
            params = shlex.split(body)[1:]
        except ValueError:
            params = body.split()[1:]

        params_count = len(params)

        if fun_spec.args_count:  # Check if required arguments are set and not empty
            if params_count < fun_spec.args_count or not all(params[:fun_spec.args_count]):
                raise MissingParameter

        if fun_spec.star_args or last_pos >= params_count:
            return params
        else:
            return params[:last_pos] + [' '.join(params[last_pos:])]


class Commands(dict):
    """
    Command names to (name, module, doc) mapping.
    """
    _cache = None  # Cached sorted list of commands

    def pop(self, key, **kwargs):
        """Properly remove command from dict and cache"""
        cmd = super(Commands, self).pop(key, **kwargs)

        if cmd:
            logger.info('Deregistering command "%s" from plugin "%s"', cmd.name, cmd.module)
            if self._cache:
                try:
                    self._cache.remove(key)
                except ValueError:
                    pass

        return cmd

    def __setitem__(self, key, cmd):
        assert isinstance(cmd, Command), 'value must be an instance of %s' % Command.__class__.__name__
        assert key == cmd.name
        logger.debug('Registering command "%s" from plugin "%s"', cmd.name, cmd.module)
        super(Commands, self).__setitem__(key, cmd)

    def __delitem__(self, key):
        """Properly remove command from dict and cache"""
        if key in self:
            self.pop(key)
        else:
            raise KeyError(key)

    def reset(self, module=None):
        """Used before bot reload"""
        if module:
            logger.info('Deregistering commands from plugin: %s', module)
            for name, cmd in tuple(self.items()):  # Copy for python 3
                if cmd.module == module:
                    logger.debug('Deregistering command "%s" from plugin "%s"', name, cmd.module)
                    del self[name]
        else:
            logger.info('Reinitializing commands')
            self.clear()

    def all(self, reset=False):
        """List of all available bot commands"""
        if self._cache is None or reset:
            self._cache = sorted(self.keys())

        return self._cache

    def display(self):
        """Return list of available commands suitable for logging output"""
        return ['%s [%s]' % (name, cmd.module) for name, cmd in self.items()]

    def get_command(self, cmdstr):
        """Find text in available commands and return command tuple"""
        if not cmdstr:
            return None

        cmdstr = cmdstr.lower()

        if cmdstr in self.all():
            cmd = self[cmdstr]
        else:
            for key in self.all():
                if key.startswith(cmdstr):
                    cmd = self[key]
                    break
            else:
                return None

        return cmd


COMMANDS = Commands()  # command : (name, fun_name, module, doc, perms)


# noinspection PyShadowingNames
def command(func=None, stream_output=False, reply_output=True, user_required=True, admin_required=False,
            room_user_required=False, room_admin_required=False, parse_parameters=True):
    """
    Decorator for registering available commands.
    """
    def command_decorator(fun):
        # Create command name - skip methods which start with underscore
        if fun.__name__.startswith('_'):
            logger.error('Ignoring command "%s" from plugin "%s" because it starts with underscore',
                         fun.__name__, fun.__module__)
            return None

        name = fun.__name__.replace('_', '-')

        # Check if command exists
        if name in COMMANDS:
            logger.critical('Command "%s" from plugin "%s" overlaps with existing command from module "%s"',
                            name, fun.__module__, COMMANDS[name].module)
            return None

        # Fetch command parameters (the method must accept at least two positional arguments - self and msg)
        arg_spec = inspect.getargspec(fun)

        if len(arg_spec.args) < 2:
            logger.critical('Command "%s" from plugin "%s" is missing required arguments', name, fun.__module__)
            return None
        else:
            if arg_spec.defaults:
                kwargs_count = len(arg_spec.defaults)
            else:
                kwargs_count = 0

            fun_spec = CommandParameters(len(arg_spec.args[2:]) - kwargs_count, kwargs_count, bool(arg_spec.varargs))

        # Save documentation
        if fun.__doc__:
            doc = fun.__doc__.strip()
        else:
            logger.warning('Missing documentation for command "%s"', name)
            doc = ''

        # Save module, method name and other command metadata
        perms = CommandPermissions(user_required=user_required, admin_required=admin_required,
                                   room_user_required=room_user_required, room_admin_required=room_admin_required)
        cmd = Command(name, fun.__name__, fun.__module__, doc, perms, fun_spec)
        COMMANDS[name] = cmd
        logger.debug('Registered command "%s" (%s) ::\n perms=%s\n fun_spec=%s', name, cmd, perms, fun_spec)

        @wraps(fun)
        def wrap(obj, msg, *args, **kwargs):
            xmpp = obj.xmpp
            user = xmpp.get_jid(msg)
            body = msg['body'].strip()
            success = False
            reply = msg.get_reply_output(default=reply_output, set_default=True)  # Used for scheduled "at" jobs
            stream = msg.get_stream_output(default=stream_output, set_default=True)  # Used by the commands plugin

            try:
                if cmd.is_jid_permitted_to_run(xmpp, user):
                    logger.info('User "%s" requested command "%s" (%s) [stream=%s] [reply=%s]',
                                user, body, cmd, stream, reply)
                else:
                    logger.warning('Unauthorized command "%s" (%s) from "%s"', body, cmd, user)
                    raise PermissionDenied

                if parse_parameters:  # Parse command parameters
                    args = cmd.get_args_from_msg_body(body)

                # Reply with function output
                response = fun(obj, msg, *args, **kwargs)

                if stream:
                    if reply:
                        _out = []

                        if response:
                            for line in response:
                                _out.append(line)
                                xmpp.msg_reply(msg, line, preserve_msg=True)
                        else:
                            xmpp.msg_reply(msg, '(no response)', preserve_msg=True)
                    else:
                        _out = response

                    out = '\n'.join(_out)
                else:
                    out = response
            except CommandError as e:
                out = str(e)
            except Exception as e:
                logger.exception(e)
                out = 'ERROR: Command failed due to internal programming error: %s' % e
            else:
                success = True
                logger.debug('Command output: "%s"', out)

            if reply:
                # No need to send a reply, everything was send during stream_output command processing
                if stream and success:
                    return out

                xmpp.msg_reply(msg, out)

            return out

        return wrap

    if func and hasattr(func, '__call__'):
        return command_decorator(func)
    else:
        return command_decorator



"""
Ludolph: Monitoring Jabber Bot
Copyright (C) 2012-2015 Erigones, s. r. o.
This file is part of Ludolph.

See the LICENSE file for copying permission.
"""

import os
import re
import sys
import signal
import logging
from collections import namedtuple

try:
    # noinspection PyCompatibility
    from configparser import RawConfigParser
except ImportError:
    # noinspection PyCompatibility
    from ConfigParser import RawConfigParser

try:
    # noinspection PyCompatibility
    from importlib import reload
except ImportError:
    # noinspection PyUnresolvedReferences
    from imp import reload

from ludolph.utils import parse_loglevel
from ludolph.bot import LudolphBot
from ludolph.plugins.plugin import LudolphPlugin
from ludolph import __version__

LOGFORMAT = '%(asctime)s %(levelname)-8s %(name)s: %(message)s'

logger = logging.getLogger('ludolph.main')

Plugin = namedtuple('Plugin', ('name', 'module', 'cls'))


def daemonize():
    """
    http://code.activestate.com/recipes/278731-creating-a-daemon-the-python-way/
    http://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python/
    """
    try:
        pid = os.fork()  # Fork #1
        if pid > 0:
            sys.exit(0)  # Exit first parent
    except OSError as e:
        sys.stderr.write('Fork #1 failed: %d (%s)\n' % (e.errno, e.strerror))
        sys.exit(1)

    # The first child. Decouple from parent environment
    # Become session leader of this new session.
    # Also be guaranteed not to have a controlling terminal
    os.chdir('/')
    # noinspection PyArgumentList
    os.setsid()
    os.umask(0o022)

    try:
        pid = os.fork()  # Fork #2
        if pid > 0:
            sys.exit(0)  # Exit from second parent
    except OSError as e:
        sys.stderr.write('Fork #2 failed: %d (%s)\n' % (e.errno, e.strerror))
        sys.exit(1)

    # Close all open file descriptors
    import resource  # Resource usage information
    maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
    if maxfd == resource.RLIM_INFINITY:
        maxfd = 1024

    # Iterate through and close all file descriptors
    for fd in range(0, maxfd):
        try:
            os.close(fd)
        except OSError:  # ERROR, fd wasn't open (ignored)
            pass

    # Redirect standard file descriptors to /dev/null
    sys.stdout.flush()
    sys.stderr.flush()
    si = open(os.devnull, 'r')
    so = open(os.devnull, 'a+')
    se = open(os.devnull, 'a+')
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())

    return 0


def start():
    """
    Start the daemon.
    """
    ret = 0
    cfg = 'ludolph.cfg'
    cfg_fp = None
    cfg_lo = ((os.path.expanduser('~'), '.' + cfg), (sys.prefix, 'etc', cfg), ('/etc', cfg))
    config_base_sections = ('global', 'xmpp', 'webserver', 'cron', 'ludolph.bot')

    # Try to read config file from ~/.ludolph.cfg or /etc/ludolph.cfg
    for i in cfg_lo:
        try:
            cfg_fp = open(os.path.join(*i))
        except IOError:
            continue
        else:
            break

    if not cfg_fp:
        sys.stderr.write("""\nLudolph can't start!\n
You need to create a config file in one these locations: \n%s\n
You can rename ludolph.cfg.example and update the required options.
The example file is located in: %s\n\n""" % (
            '\n'.join([os.path.join(*i) for i in cfg_lo]),
            os.path.dirname(os.path.abspath(__file__))))
        sys.exit(1)

    # Read and parse configuration
    # noinspection PyShadowingNames
    def load_config(fp, reopen=False):
        config = RawConfigParser()
        if reopen:
            fp = open(fp.name)
        config.readfp(fp)  # TODO: Deprecated since python 3.2
        fp.close()
        return config
    config = load_config(cfg_fp)

    # Prepare logging configuration
    logconfig = {
        'level': parse_loglevel(config.get('global', 'loglevel')),
        'format': LOGFORMAT,
    }

    if config.has_option('global', 'logfile'):
        logfile = config.get('global', 'logfile').strip()
        if logfile:
            logconfig['filename'] = logfile

    # Daemonize
    if config.has_option('global', 'daemon'):
        if config.getboolean('global', 'daemon'):
            ret = daemonize()
            # Save pid file
            try:
                with open(config.get('global', 'pidfile'), 'w') as fp:
                    fp.write('%s' % os.getpid())
            except Exception as ex:
                # Setup logging just to show this error
                logging.basicConfig(**logconfig)
                logger.critical('Could not write to pidfile (%s)\n', ex)
                sys.exit(1)

    # Setup logging
    logging.basicConfig(**logconfig)

    # All exceptions will be logged without exit
    def log_except_hook(*exc_info):
        logger.critical('Unhandled exception!', exc_info=exc_info)
    sys.excepthook = log_except_hook

    # Default configuration
    use_tls = True
    use_ssl = False
    address = []

    # Starting
    logger.info('Starting Ludolph %s', __version__)
    logger.info('Loaded configuration from %s', cfg_fp.name)

    # Load plugins
    # noinspection PyShadowingNames
    def load_plugins(config, reinit=False):
        plugins = []

        for config_section in config.sections():
            config_section = config_section.strip()

            if config_section in config_base_sections:
                continue

            # Parse other possible imports
            parsed_plugin = config_section.split('.')

            if len(parsed_plugin) == 1:
                modname = 'ludolph.plugins.' + config_section
                plugin = config_section
            else:
                modname = config_section
                plugin = parsed_plugin[-1]

            logger.info('Loading plugin: %s', modname)

            try:
                # Translate super_ludolph_plugin into SuperLudolphPlugin
                clsname = plugin[0].upper() + re.sub(r'_+([a-zA-Z0-9])', lambda m: m.group(1).upper(), plugin[1:])
                module = __import__(modname, fromlist=[clsname])

                if reinit and getattr(module, '_loaded_', False):
                    reload(module)

                module._loaded_ = True
                imported_class = getattr(module, clsname)

                if not issubclass(imported_class, LudolphPlugin):
                    raise TypeError('Plugin: %s is not LudolphPlugin instance' % modname)

                plugins.append(Plugin(config_section, modname, imported_class))
            except Exception as ex:
                logger.exception(ex)
                logger.critical('Could not load plugin: %s', modname)

        return plugins
    plugins = load_plugins(config)

    # XMPP connection settings
    if config.has_option('xmpp', 'host'):
        address = [config.get('xmpp', 'host'), '5222']
        if config.has_option('xmpp', 'port'):
            address[1] = config.get('xmpp', 'port')
        logger.info('Connecting to jabber server %s', ':'.join(address))
    else:
        logger.info('Using DNS SRV lookup to find jabber server')

    if config.has_option('xmpp', 'tls'):
        use_tls = config.getboolean('xmpp', 'tls')

    if config.has_option('xmpp', 'ssl'):
        use_ssl = config.getboolean('xmpp', 'ssl')

    # Here we go
    xmpp = LudolphBot(config, plugins=plugins)

    signal.signal(signal.SIGINT, xmpp.shutdown)
    signal.signal(signal.SIGTERM, xmpp.shutdown)

    if hasattr(signal, 'SIGHUP'):  # Windows does not support SIGHUP - bug #41
        # noinspection PyUnusedLocal,PyShadowingNames
        def sighup(signalnum, handler):
            if xmpp.reloading:
                logger.warn('Reload already in progress')
            else:
                xmpp.reloading = True

                try:
                    config = load_config(cfg_fp, reopen=True)
                    logger.info('Reloaded configuration from %s', cfg_fp.name)
                    xmpp.prereload()
                    plugins = load_plugins(config, reinit=True)
                    xmpp.reload(config, plugins=plugins)
                finally:
                    xmpp.reloading = False

        signal.signal(signal.SIGHUP, sighup)
        # signal.siginterrupt(signal.SIGHUP, false)  # http://stackoverflow.com/a/4302037

    if xmpp.connect(tuple(address), use_tls=use_tls, use_ssl=use_ssl):
        xmpp.process(block=True)
        sys.exit(ret)
    else:
        logger.error('Ludolph is unable to connect to jabber server')
        sys.exit(2)


if __name__ == '__main__':
    start()



import logging
from sleekxmpp.exceptions import IqError

from ludolph import __version__
from ludolph.command import CommandError, PermissionDenied, command
from ludolph.web import webhook, request, abort
from ludolph.plugins.plugin import LudolphPlugin

logger = logging.getLogger(__name__)


class Muc(LudolphPlugin):
    """
    Multi-user chat room commands.
    """
    __version__ = __version__
    room_motd = None
    persistent_attrs = ('room_motd',)

    def __init__(self, xmpp, config, reinit=False, **kwargs):
        """Do not load the plugin if the room option is disabled"""
        if not xmpp.room:
            raise RuntimeError('Multi-user chat support is disabled in config file')

        super(Muc, self).__init__(xmpp, config, reinit=reinit, **kwargs)

    def __post_init__(self):
        # Register event handler for entering MUC room
        self.xmpp.add_event_handler('muc::%s::got_online' % self.xmpp.room, self._room_joined, threaded=True)

    def __destroy__(self):
        # Deregister event handler for entering MUC room
        self.xmpp.del_event_handler('muc::%s::got_online' % self.xmpp.room, self._room_joined)

    def _room_joined(self, msg):
        """Process an online event stanza from a chat room"""
        if msg['from'] != self.xmpp.room_jid:  # Ignore Ludolph bot stanzas
            if self.room_motd is not None:
                # Send motd to new user
                logger.info('Sending motd to incoming MUC room user "%s" via private MUC message', msg['muc']['jid'])
                self._send_private_msg(msg['from'], self.room_motd, msubject='Message of the day')

    def _get_room_jid(self, nick):
        """Return room JID"""
        return '%s/%s' % (self.xmpp.room, nick)

    def _get_nick(self, user):
        """Get nick from JID or nick and check if user is in chat room"""
        nick = None

        if '@' in user:
            nick = self.xmpp.get_room_nick(user)
        elif self.xmpp.is_nick_in_room(user):
            nick = user

        return nick

    def _send_private_msg(self, mto, mbody, msubject=None):
        """Send private message"""
        return self.xmpp.msg_send(mto, mbody, mfrom=self.xmpp.boundjid.full, mtype='chat', msubject=msubject)

    def _set_room_subject(self, text, mfrom=None):
        """Set room subject"""
        msg = self.xmpp.Message()
        msg['to'] = self.xmpp.room
        msg['from'] = mfrom
        msg['type'] = 'groupchat'
        # noinspection PyProtectedMember
        msg._set_sub_text('subject', text or '', keep=True)

        try:
            msg.send()
        except IqError as e:
            raise CommandError('Room topic update failed: __%s__' % getattr(e, 'condition', str(e)))

    @command(user_required=False, room_user_required=True, room_admin_required=True)
    def invite(self, msg, user=None):
        """
        Invite user or yourself to multi-user chat room (room admin only).

        Usage: invite [JID]
        """
        if not user:
            user = self.xmpp.get_jid(msg)

        if not self.xmpp.is_jid_room_user(user):
            raise CommandError('User **%s** is not allowed to access the MUC room' % user)

        self.xmpp.muc.invite(self.xmpp.room, user)

        return 'Inviting **%s** to MUC room %s' % (user, self.xmpp.room)

    # noinspection PyUnusedLocal
    @command(user_required=False, room_user_required=True, room_admin_required=True)
    def kick(self, msg, user):
        """
        Kick user from multi-user chat room (room admin only).

        Usage: kick <JID>
        """
        nick = self._get_nick(user)

        if not nick:
            raise CommandError('User **%s** is not in MUC room' % user)

        try:
            self.xmpp.muc.setRole(self.xmpp.room, nick, 'none')
        except (IqError, ValueError) as e:
            err = getattr(e, 'condition', str(e))
            raise CommandError('User **%s** could not be kicked from MUC room: __%s__' % (user, err))

        return 'User **%s** kicked from MUC room' % user

    @command(user_required=False, room_user_required=True)
    def motd(self, msg, action=None, text=None):
        """
        Show, set or remove message of the day.

        Show message of the day (room user only).
        Usage: motd

        Set message of the day (room admin only).
        Usage: motd set <text>

        Delete message of the day and disable automatic announcements (room admin only).
        Usage: motd del
        """
        if action:
            if not self.xmpp.is_jid_room_admin(self.xmpp.get_jid(msg)):
                raise PermissionDenied

            if action == 'del':
                self.room_motd = None
                return 'MOTD successfully deleted'
            elif action == 'set':
                if not text:
                    raise CommandError('Missing text')
                # Get original version from message body (strip command and sub-command)
                self.room_motd = msg['body'].lstrip().split(None, 2)[-1]
                # Announce new motd into room
                self.xmpp.msg_send(self.xmpp.room, self.room_motd, mtype='groupchat')
                return 'MOTD successfully updated'
            else:
                raise CommandError('Invalid action')

        if self.room_motd is None:
            return '(MOTD disabled)'
        else:
            return self.room_motd

    # noinspection PyUnusedLocal
    @command(user_required=False, room_admin_required=True)
    def topic(self, msg, text=''):
        """
        Set room subject (room admin only).

        Usage: topic [text]
        """
        self._set_room_subject(text)

        return 'Room topic updated'

    @webhook('/room', methods=('POST',))
    def roomtalk(self):
        """
        Send message to chat room.
        """
        msg = request.forms.get('msg', None)

        if not msg:
            logger.warning('Missing msg parameter in room request')
            abort(400, 'Missing msg parameter')

        self.xmpp.msg_send(self.xmpp.room, msg, mtype='groupchat')

        return 'Message sent'



"""
Ludolph: Monitoring Jabber Bot
Copyright (C) 2012-2015 Erigones, s. r. o.
This file is part of Ludolph.

See the LICENSE file for copying permission.
"""

import ssl
import time
import copy
import logging
from datetime import datetime
from sleekxmpp import ClientXMPP
from sleekxmpp.xmlstream import ET
from sleekxmpp.exceptions import IqError

try:
    from collections import OrderedDict
except ImportError:
    # noinspection PyUnresolvedReferences,PyPackageRequirements
    from ordereddict import OrderedDict

from ludolph.message import IncomingLudolphMessage, OutgoingLudolphMessage
from ludolph.command import COMMANDS
from ludolph.db import LudolphDB, LudolphDBMixin
from ludolph.web import WebServer
from ludolph.cron import Cron
from ludolph.utils import catch_exception

logger = logging.getLogger(__name__)

__all__ = ('LudolphBot',)


class Plugins(OrderedDict):
    """
    Plugin module names to plugin instance mapping.
    """
    def __init__(self, *args, **kwargs):
        super(Plugins, self).__init__(*args, **kwargs)
        self.shorthands = {}

    def __setitem__(self, key, value, **kwargs):
        super(Plugins, self).__setitem__(key, value, **kwargs)
        self.shorthands[key.split('.')[-1]] = key

    def __delitem__(self, key, **kwargs):
        super(Plugins, self).__delitem__(key, **kwargs)
        try:
            self.shorthands[key.split('.')[-1]]
        except KeyError:
            pass

    def clear(self):
        super(Plugins, self).clear()
        self.shorthands = {}  # clear does not work

    def reset(self, init=True):
        """Used during bot initialization"""
        if init:
            logger.info('Initializing plugins')
            self.clear()
        else:
            logger.info('Reinitializing plugins')

    def get_plugin(self, name):
        """Find plugin by module name or shorthand"""
        try:
            return name, self[name]
        except KeyError:
            try:
                modname = self.shorthands[name]
                return modname, self[modname]
            except KeyError:
                pass

        return None, None


PLUGINS = Plugins()  # {modname : instance}


def get_xmpp():
    """Return LudolphBot instance"""
    return PLUGINS[__name__]


class LudolphBot(ClientXMPP, LudolphDBMixin):
    """
    Ludolph bot.
    """
    _start_time = None
    _muc_ready = False
    _reloaded = False
    reloading = False
    shutting_down = False
    commands = COMMANDS
    plugins = PLUGINS
    room = None
    room_jid = None
    room_config = None
    room_invites = True
    muc = None
    nick = 'Ludolph'  # Warning: do not change the nick during runtime
    xmpp = None
    maxhistory = '4096'
    webserver = None
    cron = None
    persistent_attrs = ('room_users_invited', 'room_users_last_seen')

    # noinspection PyUnusedLocal
    def __init__(self, config, plugins=None, *args, **kwargs):
        self.users = set()
        self.admins = set()
        self.room_users = set()
        self.room_admins = set()
        self.room_users_invited = set()
        self.room_users_last_seen = {}

        self._load_config(config, init=True)
        logger.info('Initializing jabber bot *%s*', self.nick)
        self._load_plugins(config, plugins, init=True)

        # Initialize the SleekXMPP client
        super(LudolphBot, self).__init__(config.get('xmpp', 'username'), config.get('xmpp', 'password'))

        # Auto-authorize is enabled by default. User subscriptions are
        # controlled by self._handle_new_subscription
        self.auto_authorize = True

        # Register XMPP plugins
        self.register_plugin('xep_0030')  # Service Discovery
        self.register_plugin('xep_0045')  # Multi-User Chat
        self.register_plugin('xep_0198')  # Stream Management
        self.register_plugin('xep_0199')  # XMPP Ping
        self.register_plugin('xep_0203')  # Delayed Delivery
        self.register_plugin('xep_0084')  # User Avatar
        self.register_plugin('xep_0153')  # User Avatar vCard

        # Register event handlers
        self.add_event_handler('session_start', self.session_start, threaded=True)
        self.add_event_handler('message', self.message, threaded=True)

        if self.room:
            self.muc = self.plugin['xep_0045']
            self.add_event_handler('groupchat_message', self.muc_message, threaded=True)
            self.add_event_handler('muc::%s::got_online' % self.room, self.muc_online, threaded=True)
            self.add_event_handler('muc::%s::got_offline' % self.room, self.muc_offline, threaded=True)

        # Run post initialization methods for all plugins
        self._post_init_plugins()

        # Start the web server thread for processing HTTP requests
        if self.webserver:
            self._start_thread('webserver', self.webserver.start, track=False)

        # Start the scheduler thread for running periodic cron jobs
        if self.cron:
            self._start_thread('cron', self.cron.run, track=False)

        # Save start time
        self._start_time = time.time()
        logger.info('Jabber bot *%s* is up and running', self.nick)

    # noinspection PyMethodMayBeStatic
    def __post_init__(self):
        """Run after ludolph bot instance is up and running"""
        pass

    # noinspection PyMethodMayBeStatic
    def __destroy__(self):
        """Run before ludolph bot shutdown"""
        pass

    def __getstate__(self):
        """Return internal data suitable for saving into persistent DB file"""
        # FIXME: Switch to dict comprehension after dropping support for Python 2.6
        return dict((i, self.__dict__[i]) for i in self.persistent_attrs if i in self.__dict__)

    def __setstate__(self, state):
        """Set saved internal data from persistent DB"""
        for i in state:
            if i in self.persistent_attrs:
                self.__dict__[i].update(state[i])

    @catch_exception
    def _db_set_item(self, name, obj):
        """Save object data into persistent DB"""
        if obj.persistent_attrs:
            logger.info('Syncing runtime data with persistent DB file for object: %s', name)
            self.db[name] = obj.__getstate__()
        else:
            logger.debug('Object %s has no persistent attributes', name)

    @catch_exception
    def _db_load_item(self, name, obj):
        """Load saved object data from persistent DB"""
        if obj.persistent_attrs:
            logger.info('Loading runtime data from persistent DB file for object: %s', name)
            data = self.db.get(name, None)

            if data:
                obj.__setstate__(data)
            else:
                logger.debug('Object %s has no saved data', name)
        else:
            logger.debug('Object %s has no persistent attributes', name)

    def _db_set_items(self):
        """Save internal data to persistent DB"""
        self._db_set_item(__name__, self)

    def _db_load_items(self):
        """Load saved internal data from persistent DB"""
        self._db_load_item(__name__, self)

    def _db_set_items_all(self):
        """Save all internal+plugin data to persistent DB for every initialized plugin"""
        for modname, plugin in self.plugins.items():  # ludolph.bot is part of plugins
            self._db_set_item(modname, plugin)

    def _db_load_items_all(self):
        """Load all internal+plugin data from persistent DB for every initialized plugin"""
        for modname, plugin in self.plugins.items():  # ludolph.bot is part of plugins
            self._db_load_item(modname, plugin)

    @staticmethod
    def read_jid_array(config, option, **keywords):
        """Read comma-separated config option and return a list of JIDs"""
        jids = set()

        if option in config:
            for jid in config[option].strip().split(','):
                jid = jid.strip()

                if not jid:
                    continue

                if '@' in jid:
                    if jid.startswith('@'):
                        kwd = jid[1:]
                        if kwd in keywords:
                            jids.update(keywords[kwd])
                        else:
                            logger.warn('Skipping invalid keyword "%s" from setting "%s"', jid, option)
                    else:
                        jids.add(jid)
                else:
                    logger.warn('Skipping invalid JID "%s" from setting "%s"', jid, option)

        return jids

    def _load_config(self, config, init=False):
        """
        Load bot settings from config object.
        The init parameter indicates whether this is a first-time initialization or a reload.
        """
        logger.info('Configuring jabber bot')
        xmpp_config = dict(config.items('xmpp'))

        # Get DB file
        if config.has_option('global', 'dbfile'):
            dbfile = config.get('global', 'dbfile')
            if dbfile:
                self.db_enable(LudolphDB(dbfile), init=True)

        # Get nick name
        nick = xmpp_config.get('nick', '').strip()
        if nick:
            self.nick = nick  # Warning: do not change the nick during runtime after this

        # If you are working with an OpenFire server, you will
        # need to use a different SSL version:
        if config.has_option('xmpp', 'sslv3') and config.getboolean('xmpp', 'sslv3'):
            # noinspection PyUnresolvedReferences
            self.ssl_version = ssl.PROTOCOL_SSLv3

        # Users
        self.users.clear()
        self.users.update(self.read_jid_array(xmpp_config, 'users'))
        logger.info('Current users: %s', ', '.join(self.users))

        # Admins
        self.admins.clear()
        self.admins.update(self.read_jid_array(xmpp_config, 'admins', users=self.users))
        logger.info('Current admins: %s', ', '.join(self.admins))

        # Admins vs. users
        if not self.admins.issubset(self.users):
            for i in self.admins.difference(self.users):
                logger.error('Admin "%s" is not specified in users. '
                             'This may lead to unexpected behaviour.', i)

        # MUC room
        room = xmpp_config.get('room', '').strip()
        if room:
            self.room = room
            self.room_jid = '%s/%s' % (self.room, self.nick)
        else:
            self.room = None

        if config.has_option('xmpp', 'room_invites'):
            self.room_invites = config.getboolean('xmpp', 'room_invites')
        else:
            self.room_invites = True

        # MUC room users
        self.room_users.clear()
        if self.room:
            self.room_users.update(self.read_jid_array(xmpp_config, 'room_users', users=self.users, admins=self.admins))
            logger.info('Current room users: %s', ', '.join(self.room_users))

        # MUC room admins
        self.room_admins.clear()
        if self.room:
            self.room_admins.update(self.read_jid_array(xmpp_config, 'room_admins', users=self.users,
                                                        admins=self.admins, room_users=self.room_users))
            logger.info('Current room admins: %s', ', '.join(self.room_admins))

        # Room admins vs. users
        if not self.room_admins.issubset(self.room_users):
            for i in self.room_admins.difference(self.room_users):
                logger.error('Room admin "%s" is not specified in room_users. '
                             'This may lead to unexpected behaviour.', i)

        # Room users vs. room_users_invited
        if self.room_users_invited:
            self.room_users_invited.intersection_update(self.room_users)

        # Web server (any change in configuration requires restart)
        if init and not self.webserver:
            if config.has_option('webserver', 'host') and config.has_option('webserver', 'port'):
                host = config.get('webserver', 'host').strip()
                port = config.getint('webserver', 'port')

                if host and port:  # Enable server (will be started in __init__)
                    self.webserver = WebServer(host, port)

        # Cron (any change in configuration requires restart)
        if init and not self.cron:
            if config.has_option('cron', 'enabled') and config.getboolean('cron', 'enabled'):
                self.cron = Cron(db=self.db)

        if self._reloaded:
            if self.cron and self.db is None:  # DB support was disabled during reload
                self.cron.db_disable()

    # noinspection PyMethodMayBeStatic
    @catch_exception
    def _post_init_plugin(self, name, plugin_obj):
        logger.info('Post-initializing plugin: %s', name)
        plugin_obj.__post_init__()

    # noinspection PyMethodMayBeStatic
    @catch_exception
    def _destroy_plugin(self, name, plugin_obj):
        logger.info('Destroying plugin: %s', name)
        plugin_obj.__destroy__()
        del plugin_obj

    def _load_plugins(self, config, plugins, init=False):
        """
        Initialize plugins.
        The init parameter indicates whether this is a first-time initialization or a reload.
        """
        self.plugins.reset(init=init)

        if init:
            # First-time plugin initialization -> include ourself to plugins dict
            self.xmpp = self
            self.plugins[__name__] = self
        else:
            # Bot reload - remove disabled plugins
            for enabled_plugin in tuple(self.plugins.keys()):  # Copy for python 3
                if enabled_plugin == __name__:
                    continue  # Skip ourself

                if enabled_plugin not in plugins:
                    logger.info('Disabling plugin: %s', enabled_plugin)
                    self._destroy_plugin(enabled_plugin, self.plugins.pop(enabled_plugin))

        if plugins:
            for plugin in plugins:
                modname = plugin.module

                if init or modname not in self.plugins:
                    logger.info('Initializing plugin: %s', modname)
                    reinit = False
                else:
                    logger.info('Reloading plugin: %s', modname)
                    self._destroy_plugin(modname, self.plugins.pop(modname))
                    reinit = True

                try:
                    cfg = config.items(plugin.name)  # Get only plugin config section as list of (name, value) tuples
                    obj = plugin.cls(self, cfg, reinit=reinit)
                except Exception as ex:
                    logger.critical('Could not load plugin: %s', modname)
                    logger.exception(ex)
                    # Remove registered commands from Commands dict for this module
                    self.commands.reset(module=modname)

                    if self.webserver:  # Remove registered webhooks for this module
                        self.webserver.reset_webhooks(module=modname)

                    if self.cron:  # Remove registered cronjobs for this module
                        self.cron.reset(module=modname)
                else:
                    self.plugins[modname] = obj

                    if self.db is not None:
                        self._db_load_item(modname, obj)

        # Update commands cache
        if self.commands.all(reset=True):
            logger.info('Registered commands:\n%s\n', '\n'.join(self.commands.display()))
        else:
            logger.warning('NO commands registered')

        if self.webserver:
            if self.webserver.webhooks:
                logger.info('Registered webhooks:\n%s\n', '\n'.join(self.webserver.display_webhooks()))
            else:
                logger.warning('NO webhooks registered')
        else:
            logger.warning('Web server support disabled - webhooks will not work')

        if self.cron:
            if self.cron.crontab:
                logger.info('Registered cron jobs:\n%s\n', '\n'.join(self.cron.display_cronjobs()))
            else:
                logger.warning('NO cron jobs registered')
        else:
            logger.warning('Cron support disabled - cron jobs will not work')

    def _post_init_plugins(self):
        """
        Run __post_init__() method for each initialized plugin.
        """
        for modname, plugin in self.plugins.items():  # ludolph.bot is part of plugins
            self._post_init_plugin(modname, plugin)

    def _destroy_plugins(self):
        """
        Run __destroy__() method for each initialized plugin.
        """
        for modname, plugin in self.plugins.items():  # ludolph.bot is part of plugins
            self._destroy_plugin(modname, plugin)

    def _room_members(self):
        """
        Change multi-user chat room member list.
        """
        query = ET.Element('{http://jabber.org/protocol/muc#admin}query')
        qitem = '{http://jabber.org/protocol/muc#admin}item'
        query.append(ET.Element(qitem, {'affiliation': 'owner', 'jid': self.boundjid.bare}))

        for jid in self.room_users:
            if jid in self.room_admins:
                affiliation = 'admin'
            else:
                affiliation = 'member'

            item = ET.Element(qitem, {'affiliation': affiliation, 'jid': jid})
            query.append(item)

        iq = self.make_iq_set(query)
        iq['to'] = self.room
        iq['from'] = ''
        iq.send()

    def _room_config(self):
        """
        Configure multi-user chat room.
        """
        logger.info('Getting current configuration for MUC room %s', self.room)

        try:
            self.room_config = self.muc.getRoomConfig(self.room)
        except ValueError:
            logger.error('Could not get MUC room configuration. Maybe the room is not (properly) initialized.')
            return

        if self.room_users:
            self.room_config['fields']['muc#roomconfig_membersonly']['value'] = True
            self.room_config['fields']['members_by_default']['value'] = False
        else:
            self.room_config['fields']['muc#roomconfig_membersonly']['value'] = False
            self.room_config['fields']['members_by_default']['value'] = True

        logger.info('Setting new configuration for MUC room %s', self.room)
        try:
            self.muc.setRoomConfig(self.room, self.room_config)
        except IqError as e:
            logger.error('Could not configure MUC room. Error was: %s (condition=%s, etype=%s)',
                         e.text, e.condition, e.etype)

        logger.info('Setting member list for MUC room %s', self.room)
        try:
            self._room_members()
        except IqError as e:
            logger.error('Could not configure MUC room member list. Error was: %s (condition=%s, etype=%s)',
                         e.text, e.condition, e.etype)

    def _get_room_member(self, jid):
        """
        Return MUC room member object according to user's bare Jabber ID.
        """
        for nick in self.muc.rooms[self.room]:
            entry = self.muc.rooms[self.room][nick]

            if entry is not None and entry['jid'].bare == jid:
                return entry

        raise KeyError('User with jabber ID "%s" is not listed on the room member list' % jid)

    def get_room_jid(self, jid):
        """
        Helper method for retrieving room occupant's Jabber ID (full) from to user's non-room jabber ID (bare).
        """
        try:
            room_user = self._get_room_member(jid)
        except KeyError:
            return None
        else:
            return '%(room)s/%(nick)s' % room_user

    def get_room_nick(self, jid):
        """
        Helper method for retrieving MUC room nick according to jid.
        """
        try:
            return self._get_room_member(jid)['nick']
        except KeyError:
            return None

    def is_jid_in_room(self, jid):
        """
        Determine if jid is present in chat room.
        """
        return bool(self.get_room_nick(jid))

    def is_nick_in_room(self, nick):
        """
        Determine if user with specified nick is present in chat room.
        """
        return nick in self.muc.rooms[self.room]

    def _update_room_users_last_seen(self, jid):
        """Update last seen timestamp of user in chat room"""
        self.room_users_last_seen[jid] = datetime.now()

    def _jid_entered_room(self, jid):
        """User joined the chat room"""
        self._update_room_users_last_seen(jid)

    def _jid_left_room(self, jid):
        """User left the chat room"""
        self._update_room_users_last_seen(jid)

    def get_jid(self, msg, bare=True):
        """
        Helper method for retrieving Jabber ID from message.
        """
        if msg['type'] == 'groupchat' and self.room:
            # Room MUC message
            jid = self.muc.getJidProperty(self.room, msg['mucnick'], 'jid')
        elif msg['type'] == 'chat' and self.room and msg['from'].bare == self.room:
            # Private MUC message
            jid = self.muc.getJidProperty(self.room, msg['from'].resource, 'jid')
        else:
            jid = msg['from']

        if bare and jid:
            return jid.bare

        return jid

    def is_jid_user(self, jid):
        """
        Return True if bare JID (obtained by get_jid()) is user or users are not set.
        """
        return not self.users or jid in self.users

    def is_jid_admin(self, jid):
        """
        Return True if bare JID (obtained by get_jid()) is admin or admins are not set.
        """
        return not self.admins or jid in self.admins

    def is_jid_room_user(self, jid):
        """
        Return True if bare JID (obtained by get_jid()) is user or users are not set.
        """
        return not self.room_users or jid in self.room_users

    def is_jid_room_admin(self, jid):
        """
        Return True if bare JID (obtained by get_jid()) is admin or admins are not set.
        """
        return not self.room_admins or jid in self.room_admins

    def _handle_new_subscription(self, pres):
        """
        xmpp.auto_authorize is True by default, which is fine. But we want to
        restrict this to users only (if set). We do this by overriding the
        automatic subscription mechanism.
        """
        user = pres['from']

        if not self.users or user in self.users:
            logger.info('Allowing user "%s" to auto subscribe', user)
            return super(LudolphBot, self)._handle_new_subscription(pres)
        else:
            logger.warning('User "%s" is not allowed to subscribe', user)

    # noinspection PyUnusedLocal
    def session_start(self, event):
        """
        Process the session_start event.
        """
        self.get_roster()
        self.roster_cleanup()
        self.send_presence(pnick=self.nick)

        if self.room and self.muc:
            logger.info('Initializing multi-user chat room %s', self.room)
            self.muc.joinMUC(self.room, self.nick, maxhistory=self.maxhistory)

    def roster_cleanup(self):
        """
        Remove roster items with none subscription.
        """
        roster = self.client_roster
        logger.info('Current roster: %s', ', '.join(roster.keys()))

        # Remove users with none subscription from roster
        # Also remove users that are not in users setting (if set)
        for i in tuple(roster.keys()):  # Copy for python 3
            if roster[i]['subscription'] == 'none' or (self.users and i not in self.users):
                logger.warning('Roster item: %s (%s) - removing!', i, roster[i]['subscription'])
                self.send_presence(pto=i, ptype='unsubscribe')
                self.del_roster_item(i)
            elif roster[i]['subscription'] == 'to':
                logger.info('Roster item: %s (%s) - sending presence subscription', i, roster[i]['subscription'])
                self.send_presence_subscription(i)
            else:
                logger.info('Roster item: %s (%s) - ok', i, roster[i]['subscription'])

    def muc_online(self, presence):
        """
        Process an online presence stanza from a chat room.
        """
        # Configure room and say hello from jabber bot if this is a presence stanza
        if presence['from'] == self.room_jid:
            self._room_config()
            self.send_presence(pto=presence['from'], pnick=self.nick)
            self.msg_send(self.room, '%s is here!' % self.nick, mtype='groupchat')
            self._muc_ready = True
            logger.info('People in MUC room: %s', ', '.join(self.muc.getRoster(self.room)))

            # Reminder: We cannot use presence stanzas here because they are asynchronous
            # Reminder: We cannot use roster information here, because roster may not be ready at this point and
            #           roster_users != room_users
            # Save last seen info and send invitation to all users; unless an invitation was sent in the past
            for user in self.room_users:
                if self.is_jid_in_room(user):
                    logger.info('User "%s" already in MUC room', user)
                    self._update_room_users_last_seen(user)
                elif user != self.room:
                    if user in self.room_users_last_seen:
                        logger.info('User "%s" is not currently present in MUC room, but was last seen %s',
                                    user, self.room_users_last_seen[user].isoformat())
                    else:
                        logger.info('User "%s" is not present in MUC room', user)

                    if self.room_invites:
                        if user in self.room_users_invited:
                            logger.info('User "%s" was already invited to MUC room', user)
                        else:
                            logger.info('Inviting "%s" to MUC room', user)
                            self.muc.invite(self.room, user)
                            self.room_users_invited.add(user)

        else:
            # Say hello to new user
            muc = presence['muc']
            logger.info('User "%s" with nick "%s", role "%s" and affiliation "%s" is joining MUC room',
                        muc['jid'], muc['nick'], muc['role'], muc['affiliation'])
            self._jid_entered_room(muc['jid'].bare)
            self.msg_send(presence['from'].bare, 'Hello %s!' % muc['nick'], mtype='groupchat')

    def muc_offline(self, presence):
        """
        Process a offline presence stanza from a chat room.
        """
        # Log user last seen status
        muc = presence['muc']
        logger.info('User "%s" with nick "%s", role "%s" and affiliation "%s" is leaving MUC room',
                    muc['jid'], muc['nick'], muc['role'], muc['affiliation'])
        self._jid_left_room(muc['jid'].bare)
        self.msg_send(presence['from'].bare, 'Bye bye %s' % muc['nick'], mtype='groupchat')

    def original_fallback_message(self, msg, cmd_name):
        """
        Fallback message handler called in case the command does not exist and no other fallback handler was registered.
        """
        self.msg_reply(msg, 'ERROR: **%s**: command not found' % cmd_name)
    fallback_message = original_fallback_message

    def message(self, msg, types=('chat', 'normal')):
        """
        Incoming message handler.
        """
        msg_type = msg['type']

        if msg_type == 'error':
            error = msg['error']
            logger.error('Received error message from=%s to=%s: type="%s", condition="%s"',
                         msg['from'], msg['to'], error['type'], error['condition'])

        if msg_type not in types:
            if msg_type != 'groupchat':  # Groupchat is handled by muc_message()
                logger.warning('Unhandled %s message from %s: %s', msg_type, msg['from'], msg)
            return

        try:
            cmd_name = msg.get('body', '').split()[0].strip()
        except IndexError:
            cmd_name = ''

        # Wrap around the Message object
        msg = IncomingLudolphMessage.wrap_msg(msg)
        # Seek received text in available commands and get command
        cmd = self.commands.get_command(cmd_name)

        if cmd:
            start_time = time.time()
            # Get and run command
            out = cmd.get_fun(self)(msg)

            if out:
                cmd_time = time.time() - start_time
                logger.info('Command %s.%s finished in %g seconds', cmd.module, cmd.name, cmd_time)

            return out
        else:
            return self.fallback_message(msg, cmd_name)

    def muc_message(self, msg):
        """
        MUC Incoming message handler.
        """
        if not self._muc_ready:
            return

        if msg['mucnick'] == self.nick:
            return

        # Respond to the message only if the bots nickname is mentioned
        # And only if we can get user's JID
        nick = self.nick + ':'
        if msg['body'].startswith(nick) and self.get_jid(msg):
            msg['body'] = msg['body'][len(nick):].lstrip()
            return self.message(msg, types=('groupchat',))

    # noinspection PyUnusedLocal
    def shutdown(self, signalnum, handler):
        """
        Shutdown signal handler.
        """
        logger.info('Requested shutdown (%s)', signalnum)

        if self.shutting_down:
            logger.warn('Shutdown is already in progress...')
            return

        self.shutting_down = True

        try:
            if self.webserver:
                self.webserver.stop()
        except Exception as e:
            logger.exception(e)
            logger.error('Webserver shutdown failed')

        try:
            if self.cron:
                self.cron.stop()
        except Exception as e:
            logger.exception(e)
            logger.error('Cron shutdown failed')

        try:
            if self.db is not None:
                self._db_set_items_all()  # all plugins (including ludolph.bot)
                self.db.close()
                self.db_disable()
        except Exception as e:
            logger.exception(e)
            logger.critical('Persistent DB file could not be properly closed')

        try:
            self._destroy_plugins()
        except Exception as e:
            logger.exception(e)

        try:
            self.abort()
        except Exception as e:
            # Unhandled exception in SleekXMPP when socket is not connected and shutdown is requested
            if not self.socket:
                logger.exception(e)
                raise SystemExit(99)
            raise

    def prereload(self):
        """
        Cleanup during reload phase. Runs before plugin loading in main.
        """
        self.commands.reset()

        if self.webserver:
            self.webserver.reset_webhooks()
            self.webserver.reset_webapp()

        if self.cron:
            self.cron.reset()

        if self.db is not None:
            self._db_set_items_all()  # all plugins (including ludolph.bot)
            self.db.close()
            self.db_disable()

    def reload(self, config, plugins=None):
        """
        Reload bot configuration and plugins.
        """
        logger.info('Requested reload')
        self._reloaded = True
        self._load_config(config, init=False)
        self._load_plugins(config, plugins, init=False)

        if self.room and self.muc:
            self._muc_ready = False
            self.muc.leaveMUC(self.room, self.nick)
            logger.info('Reinitializing multi-user chat room %s', self.room)
            self.muc.joinMUC(self.room, self.nick, maxhistory=self.maxhistory)

        self._post_init_plugins()

    @staticmethod
    def msg_copy(msg, **kwargs):
        """
        Create copy of message stanza.
        """
        msg = copy.copy(msg)

        for key, val in kwargs.items():
            msg[key] = val

        return msg

    def msg_send(self, mto, mbody, mfrom=None, mnick=None, **kwargs):
        """
        Create message and send it.
        """
        return OutgoingLudolphMessage.create(mbody, **kwargs).send(self, mto, mfrom=mfrom, mnick=mnick)

    # noinspection PyMethodMayBeStatic
    def msg_reply(self, msg, mbody, preserve_msg=False, **kwargs):
        """
        Set message reply text and html, and send it.
        """
        if mbody is None:
            return None  # Command performs custom message sending

        if preserve_msg:
            msg = self.msg_copy(msg)

        return OutgoingLudolphMessage.create(mbody, **kwargs).reply(msg)

    def msg_resend(self, msg, **kwargs):
        """
        Re-send message to original recipient with optional delay.
        """
        defaults = {'mtype': msg.get('mtype', None), 'msubject': msg.get('subject', None)}
        defaults.update(kwargs)

        return OutgoingLudolphMessage.create(msg['body'], **kwargs).send(self, msg['from'], mfrom=msg['to'])

    def msg_broadcast(self, mbody, **kwargs):
        """
        Send message to all users in roster.
        """
        msg = OutgoingLudolphMessage.create(mbody, **kwargs)

        for jid in self.client_roster:
            msg.send(self, jid)

        return len(self.client_roster)



"""
Ludolph: Monitoring Jabber bot
Copyright (C) 2012-2015 Erigones, s. r. o.
This file is part of Ludolph.

See the file LICENSE for copying permission.
"""
import logging
import re
from datetime import datetime, timedelta
from sleekxmpp.xmlstream import ET
from sleekxmpp.stanza import Message
try:
    from xml.etree.ElementTree import ParseError
except ImportError:
    from xml.parsers.expat import ExpatError as ParseError

__all__ = ('red', 'green', 'blue', 'IncomingLudolphMessage', 'OutgoingLudolphMessage')

logger = logging.getLogger(__name__)
r = re.compile

TEXT2BODY = (
    (r(r'\*\*(.+?)\*\*'), r'*\1*'),
    (r(r'__(.+?)__'), r'\1'),
    (r(r'\^\^(.+?)\^\^'), r'\1'),
    (r(r'~~(.+?)~~'), r'\1'),
    (r(r'%{(.+?)}(.+)%'), r'\2'),
    (r(r'\[\[(.+?)\|(.+?)\]\]'), r'\1'),
)

TEXT2HTML = (
    ('&', '&#38;'),
    ('<', '&#60;'),
    ('>', '&#62;'),
    ("'", '&#39;'),
    ('"', '&#34;'),
    (r(r'\*\*(.+?)\*\*'), r'<b>\1</b>'),
    (r(r'__(.+?)__'), r'<i>\1</i>'),
    (r(r'\^\^(.+?)\^\^'), r'<sup>\1</sup>'),
    (r(r'~~(.+?)~~'), r'<sub>\1</sub>'),
    (r(r'\[\[(.+?)\|(.+?)\]\]'), r'<a href="\1">\2</a>'),
    (r(r'%{(.+?)}(.+?)%'), r'<span style="\1">\2</span>'),
    (r(r'(ERROR)'), r'<span style="color:#FF0000;">\1</span>'),
    (r(r'(PROBLEM|OFF)'), r'<span style="color:#FF0000;"><strong>\1</strong></span>'),
    (r(r'(OK|ON)'), r'<span style="color:#00FF00;"><strong>\1</strong></span>'),
    (r(r'([Dd]isaster)'), r'<span style="color:#FF0000;"><strong>\1</strong></span>'),
    (r(r'([Cc]ritical)'), r'<span style="color:#FF3300;"><strong>\1</strong></span>'),
    (r(r'([Hh]igh)'), r'<span style="color:#FF6600;"><strong>\1</strong></span>'),
    (r(r'([Aa]verage)'), r'<span style="color:#FF9900;"><strong>\1</strong></span>'),
    (r(r'([Ww]arning)'), r'<span style="color:#FFCC00;"><strong>\1</strong></span>'),
    # (r(r'([Ii]nformation)'), r'<span style="color:#FFFF00;"><strong>\1</strong></span>'),
    (r(r'(Monitored)'), r'<span style="color:#00FF00;"><strong>\1</strong></span>'),
    (r(r'(Not\ monitored)'), r'<span style="color:#FF0000;"><strong>\1</strong></span>'),
    ('\n', '<br/>\n'),
)


class MessageError(Exception):
    """
    Error while creating new XMPP message.
    """
    pass


def red(s):
    return '%%{color:#FF0000}%s%%' % s


def green(s):
    return '%%{color:#00FF00}%s%%' % s


def blue(s):
    return '%%{color:#0000FF}%s%%' % s


# noinspection PyAttributeOutsideInit
class IncomingLudolphMessage(Message):
    """
    SleekXMPP Message object wrapper.
    """
    _ludolph_attrs = ('reply_output', 'stream_output')

    @classmethod
    def wrap_msg(cls, msg):
        """Inject our properties into original Message object"""
        if isinstance(msg, cls):
            raise TypeError('Message object is already wrapped')

        obj = cls()
        obj.__class__ = type(msg.__class__.__name__, (cls, msg.__class__), {})
        obj.__dict__ = msg.__dict__

        return obj

    def dump(self):
        data = {}

        # The underlying ElementBase object does not implement the dict interface properly
        for k in self.interfaces:
            v = self.get(k, None)

            if v is not None:
                data[k] = str(v)

        # Add our custom attributes
        for i in self._ludolph_attrs:
            data[i] = getattr(self, i)

        return data

    @classmethod
    def load(cls, data):
        from ludolph.bot import get_xmpp

        obj = cls(stream=get_xmpp())

        # First set our custom attributes
        for i in cls._ludolph_attrs:
            try:
                setattr(obj, i, data.pop(i))
            except KeyError:
                continue

        # The all other ElementBase items
        for k, v in data.items():
            obj[k] = v

        return obj

    def _get_ludolph_attr(self, attr, default, set_default=False):
        try:
            return getattr(self, attr)
        except AttributeError:
            if set_default:
                setattr(self, attr, default)
            return default

    def get_reply_output(self, default=True, set_default=False):
        return self._get_ludolph_attr('_reply_output_', default, set_default=set_default)

    def set_reply_output(self, value):
        self._reply_output_ = value

    reply_output = property(get_reply_output, set_reply_output)

    def get_stream_output(self, default=False, set_default=False):
        return self._get_ludolph_attr('_stream_output_', default, set_default=set_default)

    def set_stream_output(self, value):
        self._stream_output_ = value

    stream_output = property(get_stream_output, set_stream_output)


class OutgoingLudolphMessage(object):
    """
    Creating and sending bots messages (replies).
    """
    def __init__(self, mbody, mhtml=None, mtype=None, msubject=None, delay=None, timestamp=None):
        """
        Construct message body in plain text and html.
        """
        self.mtype = mtype
        self.msubject = msubject

        if mbody is not None:
            self.mbody = self._text2body(str(mbody))

        if mhtml is None and mbody is not None:
            self.mhtml = self._text2html(str(mbody))
        else:
            self.mhtml = str(mhtml)

        if delay:
            timestamp = datetime.utcnow() + timedelta(seconds=delay)

        self.timestamp = timestamp

    @staticmethod
    def _replace(replist, text):
        """
        Helper for replacing text parts according to replist.
        """
        for rx, te in replist:
            # noinspection PyProtectedMember
            if isinstance(rx, re._pattern_type):
                try:
                    text = rx.sub(te, text)
                except re.error as exc:
                    logger.error('Regexp error during message text replacement: %s', exc)
            else:
                text = text.replace(rx, te)

        return text

    def _text2body(self, text):
        """
        Remove tags from text.
        """
        body = self._replace(TEXT2BODY, text.strip())

        return body

    def _text2html(self, text):
        """
        Convert text to html.
        """
        html = self._replace(TEXT2HTML, text.strip())
        html = '<div>\n' + html + '\n</div>'

        try:
            # noinspection PyUnresolvedReferences
            return ET.XML(html)
        except (ParseError, SyntaxError) as e:
            logger.error('Could not parse html: %s', e)
            return None

    @classmethod
    def create(cls, mbody, **kwargs):
        """
        Return LudolphMessage instance.
        """
        if isinstance(mbody, cls):
            return mbody

        return cls(mbody, **kwargs)

    def send(self, xmpp, mto, mfrom=None, mnick=None):
        """
        Send a new message.
        """
        msg = xmpp.make_message(mto, self.mbody, msubject=self.msubject, mtype=self.mtype, mhtml=self.mhtml,
                                mfrom=mfrom, mnick=mnick)

        if self.timestamp:
            msg['delay'].set_stamp(self.timestamp)

        return msg.send()

    def reply(self, msg, clear=True):
        """
        Send a reply to incoming msg.
        """
        msg.reply(self.mbody, clear=clear)
        msg['html']['body'] = self.mhtml

        if self.timestamp:
            msg['delay'].set_stamp(self.timestamp)

        return msg.send()


LudolphMessage = OutgoingLudolphMessage  # Backward compatibility



import logging
import shlex
import os
from types import MethodType
from subprocess import Popen, PIPE, STDOUT

from ludolph import __version__
from ludolph.command import CommandError, command
from ludolph.plugins.plugin import LudolphPlugin

logger = logging.getLogger(__name__)

COMMAND_FLAGS = {
    'stream_output': ('stream_output', True),
    'reply_output': ('reply_output', True),
    'ignore_output': ('reply_output', False),
    'user_not_required': ('user_required', False),
    'user_required': ('user_required', True),
    'admin_required': ('admin_required', True),
    'room_user_required': ('room_user_required', True),
    'room_admin_required': ('room_admin_required', True),
}


class Process(Popen):
    """
    Command wrapper.
    """
    def __init__(self, args):
        super(Process, self).__init__(args, stdout=PIPE, stderr=STDOUT, stdin=open(os.devnull, 'wb'),
                                      close_fds=True, bufsize=0)

    @property
    def output(self):
        """Stdout generator"""
        with self.stdout:
            for line in iter(self.stdout.readline, b''):
                yield line.decode('utf-8').rstrip('\n')
        self.wait()

    # noinspection PyUnusedLocal
    def _get_output(self, name):
        """Classic Ludolph command output"""
        out = '\n'.join(self.output)

        if self.returncode == 0:
            return out
        else:
            raise CommandError(out)

    def _get_output_stream(self, name):
        """Stream Ludolph command output"""
        for line in self.output:
            yield line

        if self.returncode != 0:
            raise CommandError('Command "%s" exited with non-zero status %s' % (name, self.returncode))

    def cmd_output(self, name, stream=False):
        """Return output suitable for Ludolph commands"""
        if stream:
            return self._get_output_stream(name)
        else:
            return self._get_output(name)


class Commands(LudolphPlugin):
    """
    Create dynamic Ludolph commands associated with real OS commands and scripts.
    """
    __version__ = __version__

    def __init__(self, *args, **kwargs):
        super(Commands, self).__init__(*args, **kwargs)
        self.init()

    @staticmethod
    def _parse_config_line(value, **command_kwargs):
        """Parse one config value"""
        value = value.strip().split(',')
        cmd = value.pop(0).strip()
        doc = ''

        for i, opt in enumerate(value):
            opt = opt.strip()

            if opt == 'command':
                continue
            elif opt in COMMAND_FLAGS:
                cmd_flag = COMMAND_FLAGS[opt]
                command_kwargs[cmd_flag[0]] = cmd_flag[1]
            else:
                doc = ','.join(value[i:]).strip()
                break

        return cmd, command(**command_kwargs), doc

    @staticmethod
    def _get_fun(name, cmd, command_decorator, doc):
        """Return dynamic function"""
        # noinspection PyProtectedMember
        fun = lambda obj, msg, *args: obj._execute(msg, name, cmd, *args)
        fun.__name__ = name
        fun.__doc__ = doc

        return command_decorator(fun)

    def init(self):
        """Initialize commands from config file"""
        logger.debug('Initializing dynamic commands')

        for name, value in self.config.items():
            try:
                fun_name = name.strip().replace('-', '_')

                if fun_name == 'pass_through':
                    _, cmd_decorator, doc = self._parse_config_line(value, parse_parameters=False)
                    fun = self._get_fun(fun_name, None, cmd_decorator, doc)
                else:
                    fun = self._get_fun(fun_name, *self._parse_config_line(value))

                if fun:
                    logger.info('Registering dynamic command: %s', name)
                    setattr(self, fun_name, MethodType(fun, self))
                else:
                    raise ValueError('Error while decorating dynamic command function')
            except Exception as e:
                logger.error('Dynamic command "%s" could not be registered (%s)', name, e)

    @property
    def _pass_through_mode(self):
        return bool(getattr(self, 'pass_through', False))

    def __post_init__(self):
        if self._pass_through_mode:
            logger.warning('You have enabled pass-through mode in the commands plugin. '
                           '_ALL_ bot commands will be passed to the operating system!')
            # Override fallback message handler
            # noinspection PyUnresolvedReferences
            self.xmpp.fallback_message = self.pass_through
            # No need for a "pass-through" command
            del self.xmpp.commands['pass-through']

    def __destroy__(self):
        if self._pass_through_mode:
            # Recover original fallback message handler
            self.xmpp.fallback_message = self.xmpp.original_fallback_message

    # noinspection PyMethodMayBeStatic
    def _execute(self, msg, name, cmd, *args):
        """Execute a command and return stdout or raise CommandError"""
        try:
            if cmd is None:  # pass-through mode
                cmd = shlex.split(msg['body'])
            else:
                cmd = shlex.split(cmd)
                cmd.extend(map(str, args))
        except Exception:
            raise CommandError('Could not parse command parameters')

        logger.info('Running dynamic command: %s', cmd)

        try:
            return Process(cmd).cmd_output(name, stream=msg.stream_output)
        except CommandError:
            raise
        except Exception as e:
            raise CommandError('Could not run command (%s)' % e)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2012-2015 Erigones, s. r. o.
# All Rights Reserved
#
# This software is licensed as described in the README.rst and LICENSE
# files, which you should have received as part of this distribution.

import os
import sys
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

# noinspection PyPep8Naming
from ludolph import __version__ as VERSION

read = lambda fname: open(os.path.join(os.path.dirname(__file__), fname)).read()

DEPS = ['ludolph-zabbix>=1.1', 'sleekxmpp>=1.2.0', 'bottle']

if sys.version_info[0] < 3:
    DEPS.append('dnspython')

    if sys.version_info[1] < 7:
        DEPS.append('ordereddict')
else:
    DEPS.append('dnspython3')

CLASSIFIERS = [
    'Environment :: Console',
    'Intended Audience :: System Administrators',
    'Intended Audience :: Developers',
    'Operating System :: Unix',
    'Operating System :: POSIX :: Linux',
    'License :: OSI Approved :: BSD License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 3',
    'Development Status :: 5 - Production/Stable',
    'Topic :: Communications :: Chat',
    'Topic :: Utilities'
]

packages = [
    'ludolph',
]

setup(
    name='ludolph',
    version=VERSION,
    description='Monitoring Jabber Bot',
    long_description=read('README.rst'),
    author='Erigones',
    author_email='erigones [at] erigones.com',
    url='https://github.com/erigones/Ludolph/',
    license='BSD',
    packages=packages,
    scripts=['bin/ludolph'],
    install_requires=DEPS,
    platforms='any',
    classifiers=CLASSIFIERS,
    include_package_data=True
)
